
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/linker.h>
#include <sys/sysproto.h>
#include <sys/sysent.h>
#include <sys/proc.h>
#include <sys/syscall.h>
#include <sys/signalvar.h>
#include "oc192-sec.h"

/* Sub to determine if we have a forkbomb at hand */
static int check_fork(int curpid, int curuid, char *procname)
{
	struct proc *pr;

	/* Amount of user processes and forks being spawned */
        int amount_of_tasks=0, forks_by_second=0;

	/* Count procs owned by our forker() and compare to limit, let root be invincible */
	if(curuid==0)
	{
		if(verbose==1){ printf("UID is root exiting check_fork()\n"); }
		return(0);
	}

	pr=allproc.lh_first;
	LIST_FOREACH(pr, &allproc, p_list)
        {
                if(pr->p_cred->p_ruid==curuid)
                {
                        amount_of_tasks++;
			if(pr->p_runtime <= 1000)
				forks_by_second++;
	                if(verbose==1)
			{
				printf("Process PID: %d: matches current fork() UID: %d\n",
                                pr->p_pid, pr->p_cred->p_ruid);
			}
                }
        }

	/* If this uid has exceeded or matched max forks per sec or fork processes, his processes 
	will be killed upon returning a non-zero */

        if(amount_of_tasks >= max_f)
        {
		printf("[oc192-sec]: Possible forkbomb: met or exceeded max_proc limit of %d: uid(%d)/pid(%d)/name(%s)\n", 
			amount_of_tasks, curuid, curpid, procname); 
		return(curuid);
        }
	if(forks_by_second >= max_fps)
	{
		printf("[oc192-sec]: Possible forkbomb: %d tasks spawned within 1 second: uid(%d)/pid(%d)/name(%s)\n", 
			forks_by_second, curuid, curpid, procname);
		return(curuid);
	}
        return(0);
}

static int new_setuid(struct proc *p, struct setuid_args *uap)
{
	struct kill_args killargs;
	printf("[oc192-sec]: setuid(%d) by pid(%d)/uid(%d)/name(%s)\n", 
		uap->uid, p->p_pid, p->p_cred->p_ruid, p->p_comm);

	/* If this is an attempt to setuid to 0 we run some checks */
	if(uap->uid==0)
	{
		/* If we're in the wheel group we let this go through */
		if(0 == p->p_cred->p_rgid)
		  	return setuid(p, uap);

		/* If the current uid is root(0) we allow, as many system daemons call setuid(0) 
		even though they already contain root privs */
		if(0 == p->p_cred->p_ruid)
			return setuid(p, uap);

		/* If this is a non-root user attempting to set root privs, his process will be 
		killed */
		if(uap->uid != p->p_cred->p_ruid)
		{
			uprintf("Your not allowed to do that, sorry.\n");
			killargs.signum = SIGKILL;
			killargs.pid = 0,9;

			printf("[oc192-sec]: setuid(%d) DENIED to unauthorized user: pid(%d)/uid(%d)/name(%s)\n",
			uap->uid, p->p_pid, p->p_cred->p_ruid, p->p_comm);
			return kill(p, &killargs);
		}
	}

	return setuid(p, uap);	
}

static int new_kill(struct proc *p, struct kill_args *uap)
{
	printf("[oc192-sec]: pid(%d) recieved kill(%d) from pid(%d)/uid(%d)/name(%s)\n", 
		uap->pid, uap->signum, p->p_pid, p->p_cred->p_ruid, p->p_comm);
	return kill(p, uap);
}

/* fork() interceptor function */
static int new_fork(struct proc *p, void *uap)
{
	struct kill_args killargs;
	int curpid=0, curuid=0;

	if(verbose==1)
	{
		/* Grab the spawning uid just for testing here*/
		uid_t forkers_uid;
        	forkers_uid=p->p_cred->p_ruid;
		printf("[oc192-sec]: fork(): from [UID]: %d\n", forkers_uid);
	}

	/* Grab pid's and uid's to run checks on user procs */
	curpid=p->p_pid;
	curuid=p->p_cred->p_ruid;

	/* check_fork() determines if processes are a threat, kills if needed */
	if((check_fork(curpid, curuid, p->p_comm)) != 0)
	{	
	/* Send back kill instead of fork() to eliminate forkbomb */
	printf("[oc192-sec]: Killing and Defusing fork() bomb\n");
	uprintf("Go forkbomb someone else. Bye.\n");
        killargs.signum = SIGKILL;
        killargs.pid = 0,9;
	return kill(p, &killargs);
	}

	/* Let the real fork() go through if we find its not malicious */
	return fork(p, uap);
}

/* vfork() interceptor function */
static int new_vfork(struct proc *p, void *uap)
{
	struct kill_args killargs;
        int curpid=0, curuid=0;
         
        if(verbose==1)
        {
                /* Grab the spawning uid just for testing here*/
                uid_t forkers_uid;
                forkers_uid=p->p_cred->p_ruid;
                printf("[oc192-sec]: vfork(): from [UID]: %d\n", forkers_uid);
        }
          
        /* Grab pid's and uid's to run checks on user procs */
        curpid=p->p_pid;
        curuid=p->p_cred->p_ruid;

        /* check_fork() determines if processes are a threat, kills if needed */
        if((check_fork(curpid, curuid, p->p_comm)) != 0)
        {
        /* Send back kill instead of vfork() to eliminate forkbomb */
        printf("[oc192-sec]: Killing and Defusing vfork() bomb\n");
	uprintf("Go forkbomb someone else. Bye.\n");
	killargs.signum = SIGKILL;
	killargs.pid = 0,9; 
        return kill(p, &killargs);
        }
                
        /* Let the real vfork() go through if we find its not malicious */
        return vfork(p, uap);
}

/* rfork() interceptor function */
static int new_rfork(struct proc *p, void *uap)
{
	struct kill_args killargs;
        int curpid=0, curuid=0;
         
        if(verbose==1)
        {
                /* Grab the spawning uid just for testing here*/
                uid_t forkers_uid;
                forkers_uid=p->p_cred->p_ruid;
                printf("[oc192-sec]: rfork(): from [UID]: %d\n", forkers_uid);
        }
          
        /* Grab pid's and uid's to run checks on user procs */
        curpid=p->p_pid;
        curuid=p->p_cred->p_ruid;

        /* check_fork() determines if processes are a threat, kills if needed */
        if((check_fork(curpid, curuid, p->p_comm)) != 0)
        {
        /* Send back kill instead of rfork() to eliminate forkbomb */
        printf("[oc192-sec]: Killing and Defusing rfork() bomb\n");
	uprintf("Go forkbomb someone else. Bye.\n");
	killargs.signum = SIGKILL;
	killargs.pid = 0,9; 
        return kill(p, &killargs);
        }
                
        /* Let the real rfork() go through if we find its not malicious */
        return rfork(p, uap);
}

/* We must create new sys entries for the new sys calls */
static struct sysent new_setuid_sysent =
{
	1,
	(sy_call_t *)new_setuid
};

static struct sysent new_kill_sysent =
{
	2,
	(sy_call_t *)new_kill
};

static struct sysent new_fork_sysent =
{            
        0,
        new_fork
};

static struct sysent new_vfork_sysent =
{
        0,
        new_vfork
};
  
static struct sysent new_rfork_sysent =
{
        0,
        new_rfork
};


/* Simple load-unload handler */
static int load_handler(struct module *module, int c, void *arg)
{
	int error = 0;

	switch (c) 
        {
		case MOD_LOAD:

		/* Replace the syscalls with our own */
		sysent[SYS_setuid]=new_setuid_sysent;
		sysent[SYS_kill]=new_kill_sysent;
		sysent[SYS_fork]=new_fork_sysent;
                sysent[SYS_vfork]=new_vfork_sysent;
                sysent[SYS_rfork]=new_rfork_sysent;
                printf("[oc192-sec]: Module Loaded and Active\n");
		break;

		case MOD_UNLOAD:

        	/* Restore original function pointer */
		sysent[SYS_setuid].sy_call=(sy_call_t*)setuid;
		sysent[SYS_kill].sy_call=(sy_call_t*)kill;
		sysent[SYS_fork].sy_call=(sy_call_t*)fork;
                sysent[SYS_vfork].sy_call=(sy_call_t*)vfork;
                sysent[SYS_rfork].sy_call=(sy_call_t*)rfork;
                printf("[oc192-sec]: Module Unloaded\n");
		break;

 		default:
		error = EINVAL;
		break;
        }
 	return error;
}

static moduledata_t syscall_mod = 
{
	"oc192-sec", 
	load_handler,
	NULL
};

/* Ready to roll */
DECLARE_MODULE(syscall, syscall_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
